import functools
import re
from typing import Any, Tuple

from ure import (
    Base,
    Ignore,
    MatchAll,
    MatchFirst,
    MatchNtoM,
    Optional,
    Result,
    Wrap,
    regex,
    by_result,
)
from ure.extra import Integer, Number, IPv4, IPv6, IP, delimited_list

TOKEN = Base(r"[a-zA-z_][\w_]*")
LITERAL = Base(r"([\"\'/])(?P<content>.*?(?<!\\)(\\\\)*)\1(?P<mods>[ilmsux]*)")


@LITERAL.modifiers.append
def convert_to_base(base, start, result, end):
    mods = result.names["mods"].upper()
    reg = regex(
        result.names["content"],
        flags=functools.reduce(
            lambda acum, val: acum | getattr(re, val), mods[1:], getattr(re, mods[0])
        )
        if result.names["mods"]
        else 0,
    )

    return Result(Base(reg))


class FutureToken:
    pass


def reduce_infix(base, extra):
    e = extra[0]
    acum = [e[0], [base, e[1]]]
    for operator, operand in extra[1:]:
        if operator == acum[0]:
            acum[1].append(operand)
        else:
            acum = [operator, [acum[0](acum[1]), operand]]

    return acum[0](acum[1])


class Namefy(Wrap):
    def __init__(self, name, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = name
        self.modifiers.append(self._namefy)

    def _namefy(self, base, start, result, end) -> Tuple[Result, int]:
        if result:
            result.names[self.name] = result.result
        return result

    def __repr__(self) -> str:
        return f"@{self.name}:{self.wrapped}"


class ParseAction(object):
    def __init__(self, name: str, parser, fnc) -> None:
        self.parser = parser
        self.name = name
        self.fnc = fnc

    @property
    def expr(self):
        return self.parser.compile(self.name)

    def parse(self, string: str, match_all: bool = True):
        expr = self.expr
        return expr.parse(string, match_all=match_all)

    def inner_parse(self, base: str, start: int = 0):
        expr = self.expr
        return expr.inner_parse(base, start)


class Parser:
    OPEN_PARENTHESIS = Base(r"\(")
    CLOSE_PARENTHESIS = Base(r"\)")

    OPERATOR_AND = Base(r"\&")
    OPERATOR_OR = Base(r"\|")
    OPERATOR_NAME = Base(":")

    NAME = Base(r"@[a-zA-Z_]\w+")
    FUNC = Base(r"\$[a-zA-Z_]\w+")

    def __init__(self) -> None:
        self.expr = {
            "integer": Integer,
            "number": Number,
            "ipv4": IPv4,
            "ipv6": IPv6,
            "ip_address": IP,
        }
        self.tokens = {}
        self.funcs = {"delimited_list": delimited_list}

        # lets define the parser
        self.literal = Base(
            r"([\"\'/])(?P<content>.*?(?<!\\)(\\\\)*)\1(?P<mods>[ilmsux]*)",
            modifiers=[self.get_literal],
        )

    def get_or(self, base: str, start: int, result: Any, end: int) -> Tuple[Any, int]:
        return MatchFirst

    def get_and(self, base: str, start: int, result: Any, end: int) -> Tuple[Any, int]:
        return MatchAll

    def get_literal(
        self, base: str, start: int, result: Any, end: int
    ) -> Tuple[Any, int]:
        mods = result.names["mods"].upper()
        reg = regex(
            result.names["content"],
            flags=functools.reduce(
                lambda acum, val: acum | getattr(re, val),
                mods[1:],
                getattr(re, mods[0]),
            )
            if result.names["mods"]
            else 0,
        )

        return Base(reg)

    def _namefy(
        self, base: str, start: int, result: Result, end: int
    ) -> Tuple[Any, int]:
        return Namefy(result.result[0][1:], result.result[2])

    def _funcify(self, result: Result):
        func_name = result.result[0][1:]
        args = result.result[2] if isinstance(result.result[2], list) else []
        return self.funcs[func_name](*args)

    def _inner_algebra(
        self, base: str, start: int, result: Result, end: int
    ) -> Tuple[Any, int]:
        a, o, b = result.result

        return o([a, b])

    def _parentesis(
        self, base: str, start: int, result: Result, end: int
    ) -> Tuple[Any, int]:
        return result.result[1]

    @property
    @functools.lru_cache()
    def pegparser(self) -> Base:
        # lets grow this
        parser = Wrap(None)
        algebra = Wrap(None)
        namefy = Wrap(None)
        funcify = Wrap(None)

        op_or = Base(r"\|", modifiers=[self.get_or])
        op_and = Base(r"\&", modifiers=[self.get_and])

        right_on = Base(r"[\*\+\!\?]*")

        infix_operand = MatchFirst([op_or, op_and])

        literal = Base(
            r"([\"\'/])(?P<content>.*?(?<!\\)(\\\\)*)\1(?P<mods>[ilmsux]*)",
            modifiers=[self.get_literal],
        )

        token = Base(r"\w+", modifiers=[lambda b, s, r, e: self.compile(r.result)])

        tokey = MatchFirst([funcify, literal, token])

        func_arg_list = Optional(delimited_list(tokey, Base(",")))
        funcify.wrapped = MatchAll(
            [self.FUNC, Base(r"\["), func_arg_list, Base(r"]")],
            modifiers=[by_result(self._funcify)],
        )

        global_expr = MatchAll(
            [self.OPEN_PARENTHESIS, algebra, self.CLOSE_PARENTHESIS],
            modifiers=[self._parentesis],
        )

        primary = MatchAll([MatchFirst([tokey, global_expr]), right_on])

        @primary.modifiers.append
        def say_what(base, start, result, end):
            r = result
            r = result.result[0]
            for mod in result.result[1]:
                if mod == "*":
                    r = MatchNtoM(r, n=0)
                elif mod == "+":
                    r = MatchNtoM(r, n=1)
                elif mod == "?":
                    r = Optional(r)
                elif mod == "!":
                    r = Ignore(r)

            return r

        inner_name = MatchAll(
            [self.NAME, self.OPERATOR_NAME, primary], modifiers=[self._namefy]
        )

        namefy.wrapped = MatchFirst([inner_name, primary])

        inner_algebra = MatchAll(
            [namefy, infix_operand, algebra], modifiers=[self._inner_algebra]
        )

        algebra.wrapped = MatchFirst([inner_algebra, namefy])

        return algebra

    def compile(self, token: str) -> Base:
        tok = self.tokens.get(token)

        if isinstance(tok, FutureToken):
            # we are procesing this definition, and its a recursive one
            # return it.
            self.tokens[token] = Wrap(None)
            return self.tokens[token]

        if tok:
            # we know this, we have process it before, just return it.
            return tok

        # we dont know the token, lets get it from expr
        expr = self.expr[token]
        if isinstance(expr, (list, tuple)):
            k, v = expr
        else:
            k, v = expr, None

        if isinstance(k, Base):
            # someone give us a expression already, lets return that
            self.tokens[token] = k
            return self.tokens[token]

        self.tokens[token] = FutureToken()
        parsed_expr = self.pegparser.parse(k.strip()).result

        if v:
            parsed_expr.modifiers.append(v)

        if isinstance(self.tokens[token], FutureToken):
            self.tokens[token] = parsed_expr

        else:
            self.tokens[token].wrapped = parsed_expr

        return self.tokens[token]

    def peg(self, peg_expr, name=None, decorator=None):
        def binder(fnc):
            _name = name if name else fnc.__name__
            if decorator:
                fnc = decorator(fnc)

            self.expr.update({_name: (peg_expr, fnc)})

            return ParseAction(_name, self, fnc)

        return binder

    def func(self, name=None):
        def binder(fnc):
            _name = name if name else fnc.__name__
            self.funcs.update({_name: fnc})
            return fnc

        return binder

    def inline(self, name, peg_expr, fnc=None):
        assert name not in self.expr, f"name {name} already exists"
        self.expr[name] = (peg_expr, fnc)
        return ParseAction(name, self, fnc)



from ure import (
    Base,
    MatchAll,
    MatchFirst,
    MatchNtoM,
    ResultNotFoundError,
    Wrap,
    regex,
    by_result,
    Result,
)

from ipaddress import IPv4Address, IPv6Address, AddressValueError

from contextlib import suppress

### some extra modifiers


def nullify(base, start, result, end):
    return None


Integer = Base(
    regex(r"[\+-]?\d+"), modifiers=[by_result(lambda result: int(result.result))]
)
Number = Base(
    regex(r"[-+]?([0-9]*\.[0-9]+|[0-9]+)([eE]\d+)?"),
    modifiers=[
        by_result(
            # life would be so much bettter if we could use `eval`
            lambda result: (
                float if any(t in result.result.lower() for t in (".", "e")) else int
            )(result.result)
        )
    ],
)


class Between(Wrap):
    def _between(self, base, start, result, end):
        if self.lower <= result.result <= self.upper:
            return result
        raise ResultNotFoundError()

    def __init__(self, lower, upper):
        self.wrapped = Integer
        self.lower, self.upper = lower, upper
        self._modifiers = [self._between]


# network

IPv4 = Base(
    r"(25[0-5]|2[0-4][0-9]|1?[0-9]{1,2})(\.(25[0-5]|2[0-4][0-9]|1?[0-9]{1,2})){3}",
    modifiers=[by_result(lambda r: IPv4Address(r.result))],
)

IPv6 = Base(
    r"([0-9a-fA-F]{1,4}:){6}(25[0-5]|2[0-4][0-9]|1?[0-9]{1,2})(\.(25[0-5]|2[0-4][0-9]|1?[0-9]{1,2})){3}|"
    r"(([0-9a-fA-F]{1,4}(:[0-9a-fA-F]{1,4}){0,4})?::([0-9a-fA-F]{1,4}(:[0-9a-fA-F]{1,4}){0,4})?)((25[0-5]|2[0-4][0-9]|1?[0-9]{1,2})(\.(25[0-5]|2[0-4][0-9]|1?[0-9]{1,2})){3})|"
    "([0-9a-fA-F]{1,4}(:[0-9a-fA-F]{1,4}){7})|"
    "([0-9a-fA-F]{1,4}(:[0-9a-fA-F]{1,4}){0,6})?::([0-9a-fA-F]{1,4}(:[0-9a-fA-F]{1,4}){0,6})?"
)


@IPv6.modifiers.append
@by_result()
def _(result: Result) -> IPv6Address:
    print(result)  # * 1 / 0
    with suppress([ValueError, AddressValueError]):
        return IPv6Address(result.result)
    raise ResultNotFoundError(f"{result.result} is not an ipv6")


def delimited_list(expr, sep, trailing=True):
    tail_element = MatchAll([sep, expr], modifiers=[by_result(lambda r: r.result[1])])
    tail = MatchNtoM(tail_element)

    delim_list_parser = MatchAll(
        [expr, tail] + ([MatchNtoM(sep, n=0, m=1)] if trailing else []),
        modifiers=[by_result(lambda r: [r.result[0], *r.result[1]])],
    )

    return delim_list_parser


IP = MatchFirst([IPv6, IPv4])

Name = Base(regex(r"[_\w][_\w\d]*"))



import ast
import inspect
import re
from contextlib import suppress
from functools import wraps
from typing import Any, Callable, Dict, List
from typing import Optional as TOptional
from typing import Pattern, Tuple, Union

MEM_REGEX: Dict[int, Pattern] = {}


def regex(expr: str, flags: int = 0) -> Pattern:
    h = hash((expr, flags))
    if h not in MEM_REGEX:
        MEM_REGEX[h] = re.compile(expr, flags)

    return MEM_REGEX[h]


class ResultTuple:
    def __init__(self, result: Any, end: int):
        self.result = result
        self.end = end


class Result:
    def __init__(
        self,
        result: List[Any],
        names: Dict[str, Any] = None,
        start: int = None,
        end: int = None,
    ):
        self.result = result
        self.names = names or {}
        self.start, self.end = start, end

    def __repr__(self) -> str:
        return f"<Result {self.result} {self.names} [{self.start}, {self.end}) >"

    def __eq__(self, result_b):
        return (result_b.result, result_b.names, result_b.start, result_b.end) == (
            self.result,
            self.names,
            self.start,
            self.end,
        )


class ResultNotFoundError(Exception):
    pass


class Base:
    def __init__(
        self,
        expr: Union[str, Pattern],
        ws: Pattern = regex(r"\s*"),
        modifiers: TOptional[
            List[Callable[[str, int, Result, int], Tuple[Result, int]]]
        ] = None,
    ) -> None:
        self.expr = regex(expr) if isinstance(expr, str) else expr
        self._ws = ws
        self._modifiers = modifiers or []

    def __repr__(self) -> str:
        return f"/{self.expr}/"

    @property
    def ws(self) -> Pattern:
        self._ws = getattr(self, "_ws", regex(r"\s*"))
        return self._ws

    @ws.setter
    def ws(self, value) -> None:
        self._ws = value

    @property
    def modifiers(self) -> List[Callable[[str, int, Result, int], Tuple[Result, int]]]:
        self._modifiers = getattr(self, "_modifiers", [])
        return self._modifiers

    def parse(self, base: str, match_all: bool = True):
        result, end = self.inner_parse(base)
        if match_all and end != len(base):
            _, end = self.parse_ws(base, end)
            if end != len(base):
                raise ResultNotFoundError()
        return result

    def _parse_any(self, base: str, start: int, e: Pattern) -> Tuple[str, int]:
        """Match a single expr"""
        result = e.match(base, start)
        if result:
            return result, result.end()
        raise ResultNotFoundError()

    def parse_ws(self, base: str, start: int) -> Tuple[str, int]:
        ws = self.ws
        if ws:
            return None, self._parse_any(base, start, ws)[1]
        return None, start

    def parse_expr(self, base: str, start: int) -> Tuple[Result, int]:
        expr, end = self._parse_any(base, start, self.expr)
        return Result(expr.group(), names=expr.groupdict(), start=start, end=end), end

    def inner_parse(self, base: str, start: int = 0) -> Tuple[Result, int]:
        _, start = self.parse_ws(base, start)

        expr, end = self.parse_expr(base, start)
        for modifier in self.modifiers:
            mod = modifier(base, start, expr, end)
            if isinstance(mod, ResultTuple):
                result, end = mod.result, mod.end
            else:
                result = mod

            if not isinstance(result, (Result, type(None))):
                expr.result = result
            else:
                expr = result

        return expr, end


class Wrap(Base):
    def __init__(
        self,
        wrapped: Base,
        modifiers: TOptional[
            List[Callable[[str, int, Result, int], Tuple[Result, int]]]
        ] = None,
    ):
        self.wrapped = wrapped
        self._modifiers = modifiers or []

    def __repr__(self):
        return f"Wrap(...)"

    def inner_parse(self, base: str, start: int = 0) -> Tuple[Result, int]:
        expr, end = self.wrapped.inner_parse(base, start)
        for modifier in self.modifiers:
            mod = modifier(base, start, expr, end)
            if isinstance(mod, ResultTuple):
                result, end = mod.result, mod.end
            else:
                result = mod

            if not isinstance(result, (Result, type(None))):
                expr.result = result
            else:
                expr = result
        return expr, end


class Composed(Base):
    @property
    def ws(self):
        return None

    def parse_ws(self, base: str, start: int):
        return base, start


class MatchAll(Composed):
    def __init__(
        self,
        exprs: List[Base],
        modifiers: TOptional[
            List[Callable[[str, int, Result, int], Tuple[Result, int]]]
        ] = None,
    ) -> None:
        self.exprs = exprs
        self._modifiers = modifiers or []

    def __repr__(self):
        return " & ".join(f"{e}" for e in self.exprs)

    def parse_expr(self, base: str, start: int):
        result = Result([], start=start)
        for e in self.exprs:
            r, start = e.inner_parse(base, start)
            if isinstance(r, Result):
                result.result.append(r.result)
                result.names.update(r.names)
            elif r:
                result.result.append(r)

        result.end = start
        return result, start


class MatchFirst(Base):
    def __init__(
        self,
        exprs: List[Base],
        modifiers: TOptional[
            List[Callable[[str, int, Result, int], Tuple[Result, int]]]
        ] = None,
    ) -> None:
        self.exprs = exprs
        self._modifiers = modifiers or []

    def __repr__(self):
        return " | ".join(f"{e}" for e in self.exprs)

    def parse_expr(self, base: str, start: int):
        for e in self.exprs:
            with suppress(ResultNotFoundError):
                return e.inner_parse(base, start)
        raise ResultNotFoundError()


class MatchNtoM(Base):
    def __init__(
        self,
        expr: Base,
        n: int = 0,
        m=None,
        modifiers: TOptional[
            List[Callable[[str, int, Result, int], Tuple[Result, int]]]
        ] = None,
    ) -> None:
        self.expr = expr
        self.n = n
        self.m = m
        self._modifiers = modifiers or []

    def parse_expr(self, base: str, start: int):
        result = Result([], start=start)
        for _ in range(self.n):
            r, start = self.expr.inner_parse(base, start)

            result.result.append(r.result)

        i = self.n
        while i < (i + 1 if self.m is None else self.m):
            try:
                r, start = self.expr.inner_parse(base, start)
            except ResultNotFoundError:
                break
            result.result.append(r.result)
            i += 1
        result.end = start
        return result, start


class Ignore(Wrap):
    def __init__(self, wrapped):
        super().__init__(wrapped)
        self.modifiers.append(lambda base, start, result, end: None)


class Optional(MatchNtoM):
    def __init__(self, wrapped: Base, empty_on_no_result: bool = False):
        super().__init__(wrapped, 0, 1)
        if not empty_on_no_result:
            self.modifiers.append(self.optionalify)

    def optionalify(self, base, start, result, end):
        if result.result:
            result.result = result.result[0]
        else:
            result = None

        return result


def by_result(fnc=None):
    if not fnc:
        return by_result

    @wraps(fnc)
    def _(text, start, result, end):
        return fnc(result)

    return _


def by_name(fnc=None):
    if not fnc:
        return by_name

    expr_sig = inspect.signature(fnc).parameters

    @wraps(fnc)
    def _(text, start, result, end):
        return fnc(**{k: result.names[k] for k in expr_sig if k in result.names})

    return _


def extract_name(name):
    def _(text, start, result, end):
        return result.names[name]

    return _
